//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
//
//  Copyright(c)  2006  Analogic Corporation	(rjohnson@analogic.com)
//
//  This program may be distributed under the GNU Public License
//  version 2, as published by the Free Software Foundation, Inc.,
//  59 Temple Place, Suite 330 Boston, MA, 02111.
//
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
//
//  File: chs.c  created Fri Feb 17 07:37:08 EST 2006	Richard B. Johnson
//  Reads disk parameter tables.
//
//  The BIOS puts the disk translation tables in R/W memory and points
//  a vector to them after it does a "read capacity" command.
//  The two vectors are 0x41 and 0x46, located in the interrupt table.
//
//  This code memory-maps the respective areas and writes the data
//  to the screen.
//
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/mman.h>

#if !defined(PAGE_SIZE)
#define PAGE_SIZE 0x1000
#endif
#if !defined(PAGE_MASK)
#define PAGE_MASK ~(PAGE_SIZE - 1)
#endif
#if !defined(MAP_FAILED)
#define MAP_FAILED (void *) -1
#endif

#define ERRORS \
    { fprintf(stderr, errmsg, __LINE__,\
      __FILE__, errno, strerror(errno));\
      exit(EXIT_FAILURE); }

static const char errmsg[]="Error at line %d, file %s (%d) [%s]\n";
static const char mapfile[]="/dev/mem";
#define PROT (PROT_READ|PROT_WRITE)
#define FLAGS (MAP_FIXED|MAP_SHARED)
#define HINT (caddr_t)0x80000000

typedef unsigned long UL;

struct dpb {
    uint16_t cyl;		// Cylinders
    uint8_t  hds;		// Heads
    uint8_t  res[3];		// Reserved
    uint16_t wpc;		// Write precomp cylinder
    uint8_t  rsv;		// reserved
    uint8_t  ctl;		// Control bit-map
    uint8_t  rsx;		// Reserved
    uint16_t ldz;		// Landing zone
    uint8_t  spt;		// Sectors per track
    uint8_t  fill[2];
    } __attribute__((packed));

struct segs {
    uint16_t off;
    uint16_t seg;
    };

const char *control[] = {
            "Reserved bit 0 set",
            "Reserved bit 1 set",
            "Reserved bit 2 set",
            "More than 8 heads",
            "Reserved bit 4 set",
            "Defect map present",
            "Disable retries",
            "Disable retries"
            };

void show_table(int vector)
{
    int fd;
    uint8_t *cp, ctl;
    struct dpb *dpb;
    struct segs *segs;
    uint16_t seg, off;
    size_t i, j;
    caddr_t mem;
    UL addr, base, offset;
    printf("\tDisk parameter table(s) at vector 0x%02x\n", vector);
    addr = 0x00000000;
    if((fd = open(mapfile, O_RDWR)) < 0)
        ERRORS;
    if((mem = mmap(HINT, PAGE_SIZE, PROT, FLAGS, fd, addr)) == MAP_FAILED)
        ERRORS;
    cp   = (uint8_t *) mem;			// Need pointer arithmetic
    cp  += (vector * sizeof(struct segs));	// 4 bytes per table entry
    segs = (struct segs *) cp;			// Pick up complete vector
    off  = segs->off;				// Lowest word is the offset
    seg  = segs->seg;				// Highest word is the segment
    addr = ((UL)seg * 0x10) + (UL) off;		// Convert to linear
    base = addr & PAGE_MASK;			// Page boundary
    offset = addr & ~PAGE_MASK;			// Offset into the page
    if(munmap(mem, PAGE_SIZE)< 0)
        ERRORS;
    if((mem = mmap(HINT, PAGE_SIZE, PROT, FLAGS, fd, base)) == MAP_FAILED)
        ERRORS;
    cp  = (uint8_t *) mem;
    cp += offset;
    dpb = (struct dpb *) cp;
    for(i=0; i< 2; i++)
    {
        printf("Disk%u\n", i);
        printf("    Cylinders = %d\n", (int) dpb->cyl);
        printf("      Sectors = %d\n", (int) dpb->spt);
        printf("        Heads = %d\n", (int) dpb->hds);
        printf("Write precomp = %d\n", (int) dpb->wpc);
        printf(" Landing zone = %d\n", (int) dpb->ldz);
        ctl = dpb->ctl;
        for(j = 0; j < 8; j++)
        {
            if(ctl & 0x01)
                printf("\t%s\n", control[j]);
            ctl >>= 1;
        }
        dpb++;
    }
    if(munmap(mem, PAGE_SIZE)< 0)
        ERRORS;
    (void)close(fd);
}
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
int main(int args, char *argv[])
{
    show_table(0x41);
    show_table(0x46);
    return 0;
}
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
